import datetime, logging, unittest

from telegram import Update
from telegram.chat import Chat
from telegram.message import Message
from telegram.messageentity import MessageEntity
from telegram.user import User

import settings
from meetg.utils import dict_to_obj, import_string


class BotTestCase(unittest.TestCase):

    def _drop_db(self):
        for model_class in settings.model_classes:
            Model = import_string(model_class)
            Model().drop()

    def setUp(self):
        super().setUp()
        settings.log_level = logging.WARNING
        self.bot = create_mock_bot()
        self._drop_db()

    def tearDown(self):
        super().tearDown()
        self._drop_db()


class UpdaterBotMock:
    username = 'mock_username'

    def get_me(self):
        me = dict_to_obj('Me', {'username': self.username})
        return me


def create_mock_bot():
    Bot = import_string(settings.bot_class)
    bot = Bot(mock=True)
    return bot


def _parse_entities(string):
    entities = []
    if string.startswith('/'):
        entity = MessageEntity(
            type=MessageEntity.BOT_COMMAND,
            offset=0,
            length=len(string.split()[0]),
        )
        entities.append(entity)
    return entities


def create_test_message(string, bot):
    date = datetime.datetime.now()
    chat = Chat(1, 'private')
    user = User(id=1, first_name='Firstname', is_bot=False)
    entities = _parse_entities(string)
    message = Message(
        message_id=1, text=string, date=date, chat=chat, from_user=user, entities=entities,
        bot=bot,
    )
    return message


def create_chat_obj(chat_id=None, chat_type='private'):
    if chat_id is None:
        if chat_type == 'private':
            chat_id = 1
        elif chat_type in ('group', 'supergroup'):
            chat_id = -1
    chat = Chat(chat_id, chat_type)
    return chat


def create_message_obj(
        message_id=1, chat_type='private', text='spam', user_id=1, chat_id=None, bot=None,
    ):
    """
    A helper to create fake update objects for testing purposes.
    Generate a private message by default
    """
    date = datetime.datetime.now()
    chat = create_chat_obj(chat_id=chat_id, chat_type=chat_type)
    user = User(id=user_id, first_name=f'User {user_id} first name', is_bot=False)
    entities = _parse_entities(text)
    message = Message(
        message_id=message_id, text=text, date=date, chat=chat, from_user=user,
        entities=entities, bot=bot,
    )
    return message


def create_update_obj(
        update_id=1, message=None, message_id=1, chat_type='private', message_text='spam',
        user_id=1, chat_id=None, bot=None,
    ):
    """
    A helper to create fake update objects for testing purposes.
    Generate an update with a private message by default
    """
    if message is None:
        message = create_message_obj(
            message_id=message_id, chat_type=chat_type, text=message_text, user_id=user_id,
            chat_id=chat_id, bot=bot,
        )
    update_obj = Update(update_id, message=message)
    return update_obj



import pymongo

import settings
from meetg.utils import import_string
from meetg.loging import get_logger


logger = get_logger()


class AbstractStorage:
    """Any other storage must be a subclass of this class"""

    def __init__(self, db_name, table_name, host, port):
        self.db_name = db_name
        self.table_name = table_name
        self.host = host
        self.port = port

    def create(self, entry):
        raise NotImplementedError

    def update(self, pattern, update):
        raise NotImplementedError

    def update_one(self, pattern, update):
        raise NotImplementedError

    def count(self, pattern=None):
        raise NotImplementedError

    def find(self, pattern=None):
        raise NotImplementedError

    def find_one(self, pattern=None):
        raise NotImplementedError

    def delete(self, pattern):
        raise NotImplementedError

    def delete_one(self, pattern):
        raise NotImplementedError

    def drop(self):
        raise NotImplementedError


class MongoStorage(AbstractStorage):
    """Wrapper for MongoDB collection methods"""

    def __init__(self, db_name, table_name, host='localhost', port=27017):
        super().__init__(db_name, table_name, host, port)
        self.client = pymongo.MongoClient(host=host, port=port)
        self.db = getattr(self.client, db_name)
        self.table = getattr(self.db, table_name)

    def create(self, entry):
        return self.table.insert_one(entry)

    def update(self, pattern, new_data):
        return self.table.update_many(pattern, {'$set': new_data})

    def update_one(self, pattern, new_data):
        return self.table.update_one(pattern, {'$set': new_data})

    def count(self, pattern=None):
        return self.table.count(pattern)

    def find(self, pattern=None):
        return self.table.find(pattern)

    def find_one(self, pattern=None):
        return self.table.find_one(pattern)

    def delete(self, pattern):
        return self.table.delete_many(pattern)

    def delete_one(self, pattern):
        return self.table.delete_one(pattern)

    def drop(self):
        return self.db.drop_collection(self.table_name)


class BaseDefaultModel:
    settings_table_name = None
    tg_id_field = None

    def __init__(self, test=False):
        if test:
            db_name = settings.db_name_test
        else:
            db_name = settings.db_name

        table_name = getattr(settings, self.settings_table_name)
        Storage = import_string(settings.storage_class)
        self._storage = Storage(
            db_name=db_name, table_name=table_name, host=settings.db_host, port=settings.db_port,
        )

    def _validate(self, data):
        validated_data = {field: data[field] for field in data if field in self.fields}
        return validated_data

    def drop(self):
        self._storage.drop()

    def create(self, obj_id, data):
        data = self._validate(data)
        data[self.db_id_field] = obj_id
        self._storage.create(data)
        logger.info('%s %s added to DB', self.name, obj_id)
        return data

    def create_from_obj(self, obj):
        obj_id = getattr(obj, self.tg_id_field)
        data = self.create(obj_id, obj.to_dict())
        return data

    def update(self, obj_id, data):
        data = self._validate(data)
        self._storage.update_one({self.tg_id_field: obj_id}, data)
        logger.info('%s %s updated in DB', self.name, obj_id)
        data[self.db_id_field] = obj_id
        return data

    def update_from_obj(self, obj):
        obj_id = getattr(obj, tg_id_field)
        data = self.update(obj_id, obj.to_dict())
        return data

    def find(self, pattern=None):
        return [obj for obj in self._storage.find(pattern)]

    def find_one(self, obj_id):
        return self._storage.find_one({self.tg_id_field: obj_id})


class DefaultUserModel(BaseDefaultModel):
    """
    Model to save and read Users in database.
    Note that field for user.id called user_id, not id.
    Other fields have the same names as in PTB
    """
    name = 'User'
    settings_table_name = 'user_table'
    tg_id_field = 'id'
    db_id_field = 'user_id'
    fields = (
        # required
        'user_id', 'is_bot', 'first_name',
        # optional
        'last_name', 'username', 'language_code', 'can_join_groups', 'can_read_all_group_messages',
        'supports_inline_queries',
        # if the user share it
        'phone_number', 'lat', 'lon',
    )


class DefaultChatModel(BaseDefaultModel):
    """
    Field for chat.id called chat_id, not id.
    """
    name = 'Chat'
    settings_table_name = 'chat_table'
    tg_id_field = 'id'
    db_id_field = 'chat_id'
    fields = (
        # required
        'chat_id', 'type',
        # optional
        'title', 'username', 'first_name', 'last_name', 'photo', 'bio', 'description',
        'invite_link', 'pinned_message', 'permissions', 'slow_mode_delay', 'sticker_set_name',
        'can_set_sticker_set', 'linked_chat_id', 'location',
    )


class DefaultMessageModel(BaseDefaultModel):
    name = 'Message'
    settings_table_name = 'message_table'
    tg_id_field = 'message_id'
    db_id_field = 'message_id'
    fields = (
        # required
        'message_id', 'date', 'chat',
        # optional
        'from', 'sender_chat', 'forward_from', 'forward_from_chat', 'forward_from_message_id',
        'forward_signature', 'forward_sender_name', 'forward_date', 'reply_to_message', 'via_bot',
        'edit_date', 'media_group_id', 'author_signature', 'text', 'entities', 'animation',
        'audio', 'document', 'photo', 'sticker', 'video', 'video_note', 'voice', 'caption',
        'caption_entities', 'contact', 'dice', 'game', 'poll', 'venue', 'location',
        'new_chat_members', 'left_chat_member', 'new_chat_title', 'new_chat_photo',
        'delete_chat_photo', 'group_chat_created', 'supergroup_chat_created',
        'channel_chat_created', 'migrate_to_chat_id', 'migrate_from_chat_id', 'pinned_message',
        'invoice', 'successful_payment', 'connected_website', 'passport_data',
        'proximity_alert_triggered', 'reply_markup',
    )



import logging


tg_api_token = ''

db_name = ''
db_name_test = ''
user_table = 'users'
chat_table = 'chats'
message_table = 'messages'

db_host = 'localhost'
db_port = 27017
storage_class = 'meetg.storage.MongoStorage'

user_model_class = 'meetg.storage.DefaultUserModel'
chat_model_class = 'meetg.storage.DefaultChatModel'
message_model_class = 'meetg.storage.DefaultMessageModel'
model_classes = (user_model_class, chat_model_class, message_model_class)

bot_class = 'bot.MyBot'
db_class = 'meetg.storage.Database'

log_path = 'log.txt'
log_level = logging.INFO



import time

import telegram
from telegram.ext import Updater

import settings
from meetg.utils import import_string
from meetg.loging import get_logger
from meetg.testing import UpdaterBotMock, create_update_obj


logger = get_logger()


class BaseBot:
    """Common Telegram bot logic"""
    save_users = True
    save_chats = True
    save_messages = False

    def __init__(self, mock=False):
        self._is_mock = mock
        if mock:
            self._tgbot = UpdaterBotMock()
            self._tgbot_username = self._tgbot.username
        else:
            self._updater = Updater(settings.tg_api_token, use_context=True)
            self._tgbot = self._updater.bot
            self._tgbot_username = self._updater.bot.get_me().username
        self._init_handlers()
        self._init_models(test=mock)

    def _init_handlers(self):
        self._handlers = self.set_handlers()
        if not self._is_mock:
            for handler in self._handlers:
                self._updater.dispatcher.add_handler(handler)

    def set_handlers(self):
        logger.warning('No handlers found')
        return ()

    def _init_models(self, test=False):
        self.user_model = import_string(settings.user_model_class)(test=test)
        self.chat_model = import_string(settings.chat_model_class)(test=test)
        self.message_model = import_string(settings.message_model_class)(test=test)

    def _mock_process_update(self, update_obj):
        """Simulation of telegram.ext.dispatcher.Dispatcher.process_update()"""
        for handler in self._handlers:
            check = handler.check_update(update_obj)
            if check not in (None, False):
                return handler.callback(update_obj, None)

    def test_send(self, message_text: str, chat_type='private'):
        """
        Method to use in auto tests.
        Simulate sending messages to the bot
        """
        update_obj = create_update_obj(
            message_text=message_text, chat_type=chat_type, bot=self._tgbot,
        )
        return self._mock_process_update(update_obj)

    def run(self):
        self._updater.start_polling()
        logger.info('%s started', self._tgbot_username)
        self._updater.idle()

    def _save(self, update_obj):
        """
        Save all the object have to be saved according to bot attrs in DB
        """
        if self.save_users:
            user = update_obj.message.from_user
            self._save_obj(user.id, user, self.user_model)
        if self.save_chats:
            chat = update_obj.message.chat
            self._save_obj(chat.id, chat, self.chat_model)
        if self.save_messages:
            message = update_obj.message
            self._save_obj(message.message_id, message, self.message_model)

    def _save_obj(self, obj_id, obj, model):
        """
        Save an object in DB
        """
        db_obj = model.find_one(obj_id)
        if db_obj:
            model.update_from_obj(obj)
        else:
            model.create_from_obj(obj)

    def extract(self, update_obj):
        """
        Extract commonly used info from update_obj,
        save users in db if they're new, log new message
        """
        chat_id = update_obj.message.chat.id
        msg_id = update_obj.message.message_id
        user = update_obj.message.from_user
        text = update_obj.message.text
        self._save(update_obj)

        contact = update_obj.message.contact
        location = update_obj.message.location

        if contact:
            logger.info('Received contact from chat %s', chat_id)
        elif location:
            logger.info('Received location from chat %s', chat_id)
        else:
            logger.info('Received message from chat %s, text length %s', chat_id, len(text or ''))
        return chat_id, msg_id, user, text

    def _log_api_call(self, method_name, kwargs):
        chat_id = kwargs.get('chat_id')
        message_id = kwargs.get('message_id')
        text = repr(kwargs.get('text', ''))

        if method_name == 'send_message':
            logger.info('Send answer to chat %s, text length %s', chat_id, len(text))
        elif method_name == 'edit_message_text':
            logger.info('Edit message %s in chat %s', message_id, chat_id)
        elif method_name == 'delete_message':
            logger.info('Delete message %s in chat %s', message_id, chat_id)
        else:
            raise NotImplementedError

    def _mock_remember(self, method_name, method_args):
        """
        If the object of the class is a mock,
        then just remember the API method and the args going to be used
        """
        self.api_method_called = method_name
        self.api_args_used = method_args
        if 'text' in method_args:
            self.api_text_sent = method_args.get('text', '')
        return None, None

    def call_bot_api(self, method_name: str, **kwargs):
        """
        Retries and handling network and load issues
        """
        if self._is_mock:
            return self._mock_remember(method_name, kwargs)

        to_attempt = 5
        success = False
        self._log_api_call(method_name, kwargs)
        method = getattr(self._tgbot, method_name)
        chat_id = kwargs.pop('chat_id', None)

        while to_attempt > 0:
            try:
                resp = method(chat_id=chat_id, **kwargs)
                success = True
                to_attempt = 0
            except telegram.error.NetworkError as exc:
                prefix = 'Network error: '
                if 'are exactly the same as' in exc.message:
                    logger.error(prefix + '"%s". It\'s ok, nothing to do here', exc.message)
                    success = True
                    to_attempt = 0
                elif "Can't parse entities" in exc.message:
                    logger.error(prefix + '"%s". Retrying is pointless', exc.message)
                    to_attempt = 0
                else:
                    logger.error(prefix + '"%s". Waiting 2 seconds then retry', exc.message)
                    to_attempt -= 1
                    time.sleep(2)
                resp = exc.message
            except telegram.error.TimedOut as exc:
                logger.error('Timed Out. Retrying')
                resp = exc.message
                to_attempt -= 1
            except telegram.error.RetryAfter as exc:
                logger.error('It is asked to retry after %s seconds. Doing', exc.retry_after)
                resp = exc.message
                to_attempt -= 2
                time.sleep(exc.retry_after + 1)
            except telegram.error.ChatMigrated as exc:
                logger.error('ChatMigrated error: "%s". Retrying with new chat id', exc)
                resp = exc.message
                chat_id = exc.new_chat_id
                to_attempt -= 1
            except (telegram.error.Unauthorized, telegram.error.BadRequest) as exc:
                logger.error('Error: "%s". Retrying', exc)
                resp = exc.message
                to_attempt -= 2
        logger.debug('Success' if success else 'Fail')
        return success, resp

    def broadcast(self, chat_ids, body, html=False):
        for chat_id in chat_ids:
            self.send_msg(chat_id, body, html=html)
        logger.info('Broadcasted: %s', body[:79])

    def send_msg(self, chat_id, body, msg_id=None, markup=None, html=None, preview=False):
        parse_mode = telegram.ParseMode.HTML if html else None
        success, resp = self.call_bot_api(
            'send_message',
            chat_id=chat_id, text=body, reply_to_message_id=msg_id, reply_markup=markup,
            parse_mode=parse_mode, disable_web_page_preview=not preview,
        )
        return success, resp

    def edit_msg_text(self, chat_id, body, msg_id, preview=False):
        success, resp = self.call_bot_api(
            'edit_message_text',
            text=body, chat_id=chat_id, message_id=msg_id, disable_web_page_preview=not preview,
        )
        return success, resp

    def delete_msg(self, chat_id, msg_id):
        success, resp = self.call_bot_api(
            'delete_message',
            chat_id=chat_id, message_id=msg_id,
        )
        return success, resp



import os
import sys
import unittest

import settings
from meetg.utils import import_string


KNOWN_ARGS = ('run', 'test')


def run():
    Bot = import_string(settings.bot_class)
    Bot().run()


def test(src_path):
    suite = unittest.loader.TestLoader().discover(src_path)
    result = unittest.runner.TextTestRunner().run(suite)


def exec_args(argv, src_path):
    if len(argv) > 1 and argv[1] in KNOWN_ARGS:
        if argv[1] == 'run':
            run()
        if argv[1] == 'test':
            test(src_path)
    else:
        print('Available commands:', ', '.join(KNOWN_ARGS))
